/* -*-c-*-
 *
 * $Id$
 *
 * Portable bit-level manipulation macros
 *
 * (c) 1998 Straylight/Edgeware
 * (c) 2000 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * Copyright (c) 2000 Mark Wooding
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2, Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. The name of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * Instead of accepting the above terms, you may redistribute and/or modify
 * this software under the terms of either the GNU General Public License,
 * or the GNU Library General Public License, published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: bits.h,v $
 * Revision 1.2  2000/07/02 15:21:20  mdw
 * Fix licence text.
 *
 * Revision 1.1  2000/05/21 11:28:30  mdw
 * Initial check-in.
 *
 * --- Past lives (mLib) --- *
 *
 * Revision 1.4  1999/12/10 23:42:04  mdw
 * Change header file guard names.
 *
 * Revision 1.3  1999/06/20 23:31:52  mdw
 * More portability enhancements.
 *
 * Revision 1.2  1999/06/17 00:12:46  mdw
 * Improve portability for shift and rotate macros.
 *
 * Revision 1.1  1999/06/01 09:46:19  mdw
 * New addition: bit manipulation macros.
 */

#ifndef BITS_H
#define BITS_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <limits.h>
#include <stddef.h>

/*----- Decide on some types ----------------------------------------------*/

/* --- Decide on a 32-bit type --- *
 *
 * I want a type which is capable of expressing 32-bit numbers.  Because some
 * implementations have 64-bit @long@s (infinitely preferable to the abortion
 * that is @long long@), using @unsigned long@ regardless is wasteful.  So,
 * if @int@ appears to be good enough, then I'll go with that.
 */

#if UINT_MAX >= 0xffffffffu
  typedef unsigned int uint32;
#else
  typedef unsigned long uint32;
#endif

/* --- Decide on a 24-bit type --- */

#if UINT_MAX >= 0x00ffffffu
  typedef unsigned int uint24;
#else
  typedef unsigned long uint24;
#endif

/* --- Decide on 16-bit and 8-bit types --- *
 *
 * This is more for brevity than anything else.
 */

typedef unsigned short uint16;
typedef unsigned char octet;

/* --- WARNING! --- *
 *
 * Never lose sight of the fact that the above types may be wider than the
 * names suggest.  Some architectures have 32-bit @short@s for example.
 */

/*----- Macros ------------------------------------------------------------*/

/* --- Useful masks --- */

#define MASK8 0xffu
#define MASK16 0xffffu
#define MASK24 0xffffffu
#define MASK32 0xffffffffu

/* --- Type coercions --- */

#define U8(x) ((octet)((x) & MASK8))
#define U16(x) ((uint16)((x) & MASK16))
#define U24(x) ((uint24)((x) & MASK24))
#define U32(x) ((uint32)((x) & MASK32))

/* --- Safe shifting macros --- */

#define LSL8(v, s) U8(U8(v) << ((s) & 7u))
#define LSR8(v, s) U8(U8(v) >> ((s) & 7u))
#define LSL16(v, s) U16(U16(v) << ((s) & 15u))
#define LSR16(v, s) U16(U16(v) >> ((s) & 15u))
#define LSL24(v, s) U24(U24(v) << ((s) % 24u))
#define LSR24(v, s) U24(U24(v) >> ((s) % 24u))
#define LSL32(v, s) U32(U32(v) << ((s) & 31u))
#define LSR32(v, s) U32(U32(v) >> ((s) & 31u))

/* --- Rotation macros --- */

#define ROL8(v, s) (LSL8((v), (s)) | (LSR8((v), 8u - (s))))
#define ROR8(v, s) (LSR8((v), (s)) | (LSL8((v), 8u - (s))))
#define ROL16(v, s) (LSL16((v), (s)) | (LSR16((v), 16u - (s))))
#define ROR16(v, s) (LSR16((v), (s)) | (LSL16((v), 16u - (s))))
#define ROL24(v, s) (LSL24((v), (s)) | (LSR24((v), 24u - (s))))
#define ROR24(v, s) (LSR24((v), (s)) | (LSL24((v), 24u - (s))))
#define ROL32(v, s) (LSL32((v), (s)) | (LSR32((v), 32u - (s))))
#define ROR32(v, s) (LSR32((v), (s)) | (LSL32((v), 32u - (s))))

/* --- Storage and retrieval --- */

#define GETBYTE(p, o) (((octet *)(p))[o] & MASK8)
#define PUTBYTE(p, o, v) (((octet *)(p))[o] = U8((v)))

#define LOAD8(p) (GETBYTE((p), 0))
#define STORE8(p, v) (PUTBYTE((p), 0, (v)))

#define LOAD16_B(p)							\
  (((uint16)GETBYTE((p), 0) << 8) |					\
   ((uint16)GETBYTE((p), 1) << 0))
#define LOAD16_L(p)							\
  (((uint16)GETBYTE((p), 0) << 0) |					\
   ((uint16)GETBYTE((p), 1) << 8))
#define LOAD16(p) LOAD16_B((p))

#define STORE16_B(p, v)							\
  (PUTBYTE((p), 0, (uint16)(v) >> 8),					\
   PUTBYTE((p), 1, (uint16)(v) >> 0))
#define STORE16_L(p, v)							\
  (PUTBYTE((p), 0, (uint16)(v) >> 0),					\
   PUTBYTE((p), 1, (uint16)(v) >> 8))
#define STORE16(p, v) STORE16_B((p), (v))

#define LOAD24_B(p)							\
  (((uint24)GETBYTE((p), 0) << 16) |					\
   ((uint24)GETBYTE((p), 1) <<  8) |					\
   ((uint24)GETBYTE((p), 2) <<  0))
#define LOAD24_L(p)							\
  (((uint24)GETBYTE((p), 0) <<  0) |					\
   ((uint24)GETBYTE((p), 1) <<  8) |					\
   ((uint24)GETBYTE((p), 2) << 16))
#define LOAD24(p) LOAD24_B((p))

#define STORE24_B(p, v)							\
  (PUTBYTE((p), 0, (uint24)(v) >> 16),					\
   PUTBYTE((p), 1, (uint24)(v) >>  8),					\
   PUTBYTE((p), 2, (uint24)(v) >>  0))
#define STORE24_L(p, v)							\
  (PUTBYTE((p), 0, (uint24)(v) >>  0),					\
   PUTBYTE((p), 1, (uint24)(v) >>  8),					\
   PUTBYTE((p), 2, (uint24)(v) >> 16))
#define STORE24(p, v) STORE24_B((p), (v))

#define LOAD32_B(p)							\
  (((uint32)GETBYTE((p), 0) << 24) |					\
   ((uint32)GETBYTE((p), 1) << 16) |					\
   ((uint32)GETBYTE((p), 2) <<  8) |					\
   ((uint32)GETBYTE((p), 3) <<  0))
#define LOAD32_L(p)							\
  (((uint32)GETBYTE((p), 0) <<  0) |					\
   ((uint32)GETBYTE((p), 1) <<  8) |					\
   ((uint32)GETBYTE((p), 2) << 16) |					\
   ((uint32)GETBYTE((p), 3) << 24))
#define LOAD32(p) LOAD32_B((p))

#define STORE32_B(p, v)							\
  (PUTBYTE((p), 0, (uint32)(v) >> 24),					\
   PUTBYTE((p), 1, (uint32)(v) >> 16),					\
   PUTBYTE((p), 2, (uint32)(v) >>  8),					\
   PUTBYTE((p), 3, (uint32)(v) >>  0))
#define STORE32_L(p, v)							\
  (PUTBYTE((p), 0, (uint32)(v) >>  0),					\
   PUTBYTE((p), 1, (uint32)(v) >>  8),					\
   PUTBYTE((p), 2, (uint32)(v) >> 16),					\
   PUTBYTE((p), 3, (uint32)(v) >> 24))
#define STORE32(p, v) STORE32_B((p), (v))

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
